/** Object to find all differences between target and actual TableOfVars.toJSON's result */
function VarsDiff(
    target,         // JSON representing wanted state
    actual,         // JSON representing actual state
    fktNameMemory,  // Function like ("Internal memory address") => "Human readable memory address" used to name memory locations
    mappings        // Optional: Allready calculated mapping from memory pointers of actual to memory pointers of target (corresponding instanced have to be removed from target.instances and actual.instances)
) {
    target = (target instanceof Array) ? { scopes: [], instances: [], vars: target } : target;
    actual = (actual instanceof Array) ? { scopes: [], instances: [], vars: actual } : actual;
    this._target = target.instances.filter(instance => ((instance.instance != 'object') || (instance.className !== null)));
    this._actual = actual.instances.filter(instance => ((instance.instance != 'object') || (instance.className !== null)));
    this.fktNameMemory = fktNameMemory;
    this._mappings = (mappings instanceof Object) ? Object.create(mappings) : {};
    this._firstReferencing = {};
    this._diffs = [];

    this._comparePart(undefined, target.vars, actual.vars);
    this._compareScopes(target.scopes, actual.scopes);
    this._compareUnusedInstances();
}

/** JSON representing wanted instances */
VarsDiff.prototype._target = null;

/** JSON representing actual instances */
VarsDiff.prototype._actual = null;

/** Function like ("Internal memory address") => "Human readable memory address" used to name memory locations */
VarsDiff.prototype.fktNameMemory = ((s) => s);

/** Calculated mapping from memory pointers of actual to memory pointers of target */
VarsDiff.prototype._mappings = {};

/** Functions creating human readable description of first corresponding variable referencing given memory location specified by target data structure */
VarsDiff.prototype._firstReferencing = {};

/** Found differences */
VarsDiff.prototype._diffs = [];

/**
* Creates a human readable description of a variable
* @return String    Created description
*/
VarsDiff.describeVar = function(
    key,            // Variable name
    descPart,       // Human readable description of part containing variable named `key` (or undefined)
    defined,        // Optional: Array of all previous variable names in present part (including `key`)
    useGenetive     // Optional: Set true if result is aimed to be used as genetive
) {
    let s = "Variable namens „" + key + '“';
    const overloads = (defined instanceof Array) ? (defined.filter(item => item == key).length - 1) : 0;
    const pres = ['', 'zweite', 'dritte', 'vierte', 'fünfte', 'sechste', 'siebte', 'achte', 'neunte', 'zehnte', 'elfte', 'zwölfte'].map(s => s ? (s + (useGenetive ? 'n ' : ' ')) : s);
    s = ((overloads < pres.length) ? pres[overloads] : (overloads + '. ')) + s;
    if (descPart)
        s += ' in ' + descPart;
    return s;
}

/**
* Compare values of variables (no references)
* @return undefined
*/
VarsDiff.prototype._compareValues = function(
    targetValue,    // Wanted value
    actualValue,    // Actual value
    getName         // Get human readable description of variable in question: (genetive) => 'Description'
) {
    if (actualValue != targetValue)
        return this._diffs.push('Wert der ' + getName(true) + ' ist falsch');
}

/**
* Compare references of variables
* @return undefined
*/
VarsDiff.prototype._compareRefs = function(
    targetRef,      // Internal memory location in target description data
    actualRef,      // Corresponding internal memory location in actual state's description data
    getName,        // Get human readable description of variable in question: (genetive) => 'Description'
    isRefToArray,   // Optional: Is targetRef a reference to an array?
    isScopes        // Optional: Are targetRef and actualRef pointer values of scopes?
) {
    if ((targetRef === undefined) && (actualRef === undefined))
        return;
    if ((actualRef === null) != (targetRef === null))
        return this._diffs.push(isScopes
            ? ('der ' + getName({ ref: actualRef }, '') + ' ' + ((actualRef === null) ? 'kann nicht' : 'müsste') + ' „this = null“ haben')
            : ('Wert der ' + getName(true) + ((targetRef === null) ? ' müsste' : ' kann nicht mehr') + ' „null“ sein'));

    if (targetRef === null)
        return;
    if (Number.isFinite(this._mappings[actualRef]) && (this._mappings[actualRef] == targetRef))
        return;

    if (Number.isFinite(this._mappings[actualRef]))
        return this._diffs.push(isScopes
            ? ('beim ' + getName({ ref: actualRef }, 'n') + ' ist der Wert von „this“ falsch')
            : (getName(false) + ' referenziert das falsche ' + (isRefToArray ? 'Array' : 'Objekt')));
    this._checkMapping(targetRef, actualRef, () => (isScopes ? getName({ ref: actualRef }, '') : getName(false)));
}

/**
* Compare two list of variables
* @return undefined
*/
VarsDiff.prototype._comparePart = function(
    descPart,       // Human readable description of part (or undefined or function(key, defined, [genetive]) { return 'description of variable named `key`'; })
    target,         // Array of variables representing wanted state
    actual          // Array of variables representing actual state
) {
    const defined = [];
    const fktNameVar = (({ key }, genetive) => ((descPart instanceof Function) ? descPart(key, defined, genetive) : VarsDiff.describeVar(key, descPart, defined, genetive)));
    actual = actual.slice();
    target.forEach(variable => {
        defined.push(variable.key);
        const actualIndex = actual.findIndex(({ key }) => key == variable.key);
        if (actualIndex == -1)
            return this._diffs.push(fktNameVar(variable, descPart) + ' fehlt');

        const actualVar = actual[actualIndex];
        actual.splice(actualIndex, 1);

        if ((actualVar.type != variable.type) || (actualVar.dimensions != variable.dimensions))
            return this._diffs.push(fktNameVar(variable, descPart) + ' müsste vom Typ ' + variable.type + '[]'.repeat(variable.dimensions) + ' sein');

        const getName = ((genetive) => fktNameVar(variable, descPart, genetive));
        return (this._compareValues(variable.value, actualVar.value, getName) || this._compareRefs(variable.ref, actualVar.ref, getName, variable.dimensions));
    });

    actual.forEach(actualVar => {
        defined.push(actualVar.key);
        this._diffs.push(fktNameVar(actualVar, descPart) + ' dürfte es nicht geben');
    });
}

/**
* After identifing a new mapping of instances: Check if instances are similar, remember mapping and remove instances from array of unmapped instances
* @return undefined
*/
VarsDiff.prototype._checkMapping = function(
    targetRef,          // Internal memory location in target description data
    actualRef,          // Corresponding internal memory location in actual state's description data
    refVarDesc          // Function to create human readable description of variable referencing
) {
    const targetIndex = this._target.map(({ pointer }) => pointer).indexOf(targetRef);
    const actualIndex = this._actual.map(({ pointer }) => pointer).indexOf(actualRef);
    if (targetIndex == -1) {
        const descFirst = this._firstReferencing[targetRef];
        this._diffs.push(refVarDesc() + ' müsste Dasselbe referenzieren wie ' + ((descFirst instanceof Function) ? descFirst() : descFirst));
        return;
    }
    const targetInstance = this._target[targetIndex];
    this._target.splice(targetIndex, 1);
    if (actualIndex == -1) {
        this._diffs.push(refVarDesc() + ' referenziert etwas, was nicht existiert…');
        return;
    }
    const actualInstance = this._actual[actualIndex];
    this._actual.splice(actualIndex, 1);
    this._mappings[actualRef] = targetRef;
    this._firstReferencing[targetRef] = refVarDesc;
    return this._compareInstancelikes(targetInstance, actualInstance);
}

/**
* Compare two instances
* @return undefined
*/
VarsDiff.prototype._compareInstancelikes = function(
    targetInstance,     // Description of wanted instance
    actualInstance      // Description of actual instance
) {
    if (targetInstance.instance != actualInstance.instance) {
        const address = this.fktNameMemory(actualInstance.pointer);
        const targetType = (targetInstance.instance == 'array') ? 'Array' : 'Objekt';
        const actualType = (targetInstance.instance == 'array') ? 'Array' : 'Objekt';
        this._diffs.push(address + ' müsste statt ein ' + actualType + ' ein ' + targetType + ' sein');
        return;
    }
    if (targetInstance.instance == 'array')
        return this._compareArrays(targetInstance, actualInstance);
    return this._compareObjects(targetInstance, actualInstance);
}

/**
* Compare two object instances
* @return undefined
*/
VarsDiff.prototype._compareObjects = function(
    targetInstance,     // Description of wanted instance
    actualInstance      // Description of actual instance
) {
    if (targetInstance.className != actualInstance.className) {
        const address = this.fktNameMemory(actualInstance.pointer);
        this._diffs.push(address + ' müsste ein Objekt der Klasse ' + targetInstance.className + ' sein');
        return;
    }
    return this._comparePart((key, defined, genetive) => {
        let desc = (genetive ? 'Attributs' : 'Attribut') + ' „' + key + '“';
        desc += ' vo' + (genetive ? 'm' : 'n') + ' Objekt ' + this.fktNameMemory(actualInstance.pointer);
        const overloads = defined.filter(item => item == key).length - 1;
        if (!overloads)
            return desc;
        const times = ['', '', 'zweifach ', 'dreifach ', 'vierfach ', 'fünffach ', 'sechsfach ', 'siebenfach ', 'achtfach ', 'neunfach ', 'zehnfach ', 'elffach ', 'zwölffach '];
        desc = (genetive ? 'überladenem ' : 'überladenes ') + desc;
        if (overloads < times.length)
            return times[overloads] + desc;
        return overloads + '-fach ' + desc;
    }, targetInstance.atts, actualInstance.atts);
}

/**
* Compare two array instances
* @return undefined
*/
VarsDiff.prototype._compareArrays = function(
    targetInstance,     // Description of wanted instance
    actualInstance      // Description of actual instance
) {
    if ((targetInstance.baseType != actualInstance.baseType) || (targetInstance.dimensions != actualInstance.dimensions)) {
        const address = this.fktNameMemory(actualInstance.pointer);
        this._diffs.push(address + ' müsste ein Array vom Typ ' + targetInstance.baseType + '[]'.repeat(targetInstance.dimensions) + ' sein');
        return;
    }
    if ((targetInstance.values || targetInstance.refs).length != (actualInstance.values || actualInstance.refs).length) {
        const address = this.fktNameMemory(actualInstance.pointer);
        this._diffs.push('die Anzahl an Einträgen des Arrays ' + address + ' müsste ' + (targetInstance.values || targetInstance.refs).length + ' sein');
        return;
    }

    const getGetName = ((index) => (genetive => ((genetive ? 'Eintrags ' : 'Eintrag ') + index + (genetive ? ' vom Array ' : ' von Array ') + this.fktNameMemory(actualInstance.pointer))));

    if (targetInstance.values instanceof Array)
        return targetInstance.values.map((targetValue, index) => this._compareValues(targetValue, actualInstance.values[index], getGetName(index))).reduce((a, b) => (a || b), null);
    if (targetInstance.refs instanceof Array)
        return targetInstance.refs
            .map((targetRef, index) => this._compareRefs(targetRef, actualInstance.refs[index], getGetName(index), targetRef ? (targetRef.dimensions > 1) : undefined))
            .reduce((a, b) => (a || b), null);
}

/**
* Compare all scopes
* @return undefined
*/
VarsDiff.prototype._compareScopes = function(
    targetScopes,
    actualScopes
) {
    if (targetScopes.length > actualScopes.length) {
        const className = targetScopes[targetScopes.length - 1].className;
        if (className === null)
            return this._diffs.push('der Methodenaufruf war noch nicht abgeschlossen');
        return this._diffs.push('das Erstellen des Objektes der Klasse „' + className + '“ ist noch nicht abgeschlossen');
    }

    if (targetScopes.length < actualScopes.length) {
        const className = actualScopes[actualScopes.length - 1].className;
        if (className === null)
            return this._diffs.push('der Methodenaufruf muss abgeschlossen sein');
        return this._diffs.push('das Erstellen des Objektes der Klasse „' + className + '“ ist abgeschlossen');
    }

    const countTypes = [0, 0];
    const nameScope = (({ className }, suf) => {
        if (className !== null)
            return 'Entwurf des Objektes der Klasse „' + className + '“';

        const occurrences = countTypes[1];
        const numberWords = ['', 'erste', 'zweite', 'dritte', 'vierte', 'fünfte', 'sechste', 'siebte', 'achte', 'neunte', 'zehnte', 'elfte', 'zwölfte'];
        const desc = 'aktiven Methodenaufruf';
        if (occurrences >= numberWords.length)
            return occurrences + '. ' + desc;
        return numberWords[occurrences] + suf + ' ' + desc;
    });

    targetScopes.forEach((targetScope, index) => {
        const actualScope = actualScopes[index];
        ++countTypes[(actualScope.className === null) ? 1 : 0];
        return this._comparePart(nameScope(actualScope, 'm'), targetScope.vars, actualScope.vars);
    });
}

/**
* Recursive compare unused instances
* @return undefined
*/
VarsDiff.prototype._compareUnusedInstances = function() {
    if (!this._target.length)
        return;
    const targetInstance = this._target[0];
    const actualCandidates = this._actual.filter(instance => ['instance', 'className', 'baseType', 'dimensions'].reduce((prev, key) => prev && (targetInstance[key] == instance[key]), true));
    if (!actualCandidates.length) {
        this._diffs.push('ein ' + ((targetInstance.instance == 'array') ? ('Array vom Typ ' + targetInstance.baseType + '[]'.repeat(targetInstance.dimensions)) : ('Objekt der Klasse ' + targetInstance.className)) + ' fehlt');
        this._target.shift();
        return this._compareUnusedInstances();
    }
    const bestCandidates = (actualCandidates.length == 1)
        ? actualCandidates
        : actualCandidates
            .map(candidate => ({
                candidate,
                diffs: (new VarsDiff(
                    { instances: this._target, vars: [{ key: 'ref', ref: targetInstance.pointer }], scopes: [] },
                    { instances: this._actual, vars: [{ key: 'ref', ref: candidate.pointer }], scopes: [] },
                    s => s,
                    this._mappings
                ))._diffs.length
            })).sort((a, b) => (a.diffs - b.diffs))
            .map(({ candidate }) => candidate);
    const actualInstance = bestCandidates[0];
    return this._checkMapping(targetInstance.pointer, actualInstance.pointer, () => VarsDiff.nameInstance(actualInstance.instance) + ' ' + this.fktNameMemory(actualInstance.pointer))
        || this._compareInstancelikes(targetInstance, actualInstance)
        || this._compareUnusedInstances();
}

/**
* Ensure `str` starts with a capital letter
* @return String        `str` with capitalized first letter
*/
VarsDiff.ucFirst = function(
    str         // String to capitalize first letter
) {
    return str.substring(0, 1).toUpperCase() + str.substring(1);
}

/**
* Get human readable description of an instance's type
* @return String        Human readable description
*/
VarsDiff.nameInstance = function(
    type        // "array" or "object"
) {
    if (type == "array")
        return "Array";
    return "Objekt";
}

/**
* Get all differences found
* @return Array         Descriptions of each found difference
*/
VarsDiff.prototype.getDiffs = function(
    upperCaseFirstLetter,           // Ensure each description start with a capital letter?
    allowAdditionalInstances        // Not complain about additional instances?
) {
    const diffs = allowAdditionalInstances ? this._diffs : this._diffs.concat(this._actual.map(instance => (
        VarsDiff.nameInstance(instance.instance) + ' ' + this.fktNameMemory(instance.pointer) + ' kann es nicht (mehr) geben'
    )));
    return upperCaseFirstLetter ? diffs.map(VarsDiff.ucFirst) : diffs.slice();
}
